/*
 * $Id: soft.c,v 1.15 1998/04/27 18:25:40 kangasha Exp kangasha $
 *
 * DEBUG: section 44    Soft Recoder
 * AUTHORS: Fabio Carignano, Young Gap Kwon, Jussi Kangasharju
 *
 * SQUID Internet Object Cache  http://squid.nlanr.net/Squid/
 * --------------------------------------------------------
 *
 *  Squid is the result of efforts by numerous individuals from the
 *  Internet community.  Development is led by Duane Wessels of the
 *  National Laboratory for Applied Network Research and funded by
 *  the National Science Foundation.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *  
 */

#ifdef _SOFT_CACH_

#include "squid.h"

#define SECTION 44               /* Debug section */
#define STORE_LOG_RECODESTART 4  /* for logging information to store.log */
#define STORE_LOG_RECODEEND 5

int soft_count = 0;              /* how many objects have been recoded */
struct _softStats SoftStats;

/* The following are internal to store.c but now they have been made visible
   to external files since we need them here in order to make the changes
   to store.c smaller. */
extern HashID store_table;
extern int storeEntryLocked _PARAMS((const StoreEntry *));
extern char *storeSwapFullPath _PARAMS((int, char *));
extern void storeSetPrivateKey _PARAMS((StoreEntry *));
extern void storeLog _PARAMS((int, const StoreEntry *));

/* Internal variables */

/* Table holding information about active recoders. */
static softrecoder_t **recoder_table = NULL;

/* How many recoders are running. */
int NRecodersAlloc = 0;

/* Internal functions */
void softStartRecoders _PARAMS((void));
int softForkRecoder _PARAMS((const char *));
void softGetRecodedResult _PARAMS((int, softrecoder_t *));
void softUpdate _PARAMS((StoreEntry *, int, int, int));

int
softForkRecoder(const char *prg)
{
    pid_t pid;
    struct sockaddr_in S;
    int cfd, sfd, fd, len;
    LOCAL_ARRAY(char, buf, 128);

    cfd = comm_open(SOCK_STREAM, 0, local_addr, 0, COMM_NOCLOEXEC,
                    "socket to recoder");
    if (cfd == COMM_ERROR) {
        debug(SECTION, 0, "softForkRecoder: Failed to create recoder\n");
        return -1;
    }
 
    len = sizeof(S);
    memset(&S, '\0', len);
    if (getsockname(cfd, (struct sockaddr *) &S, &len) < 0) {
        debug(50, 0, "softForkRecoder: getsockname: %s\n", xstrerror());
        comm_close(cfd);
        return -1;
    }
 
    comm_listen(cfd);
    if ((pid = fork()) < 0) {
	debug(50, 0, "softForkRecoder: fork: %s\n", xstrerror());
	comm_close(cfd);
	return -1;
    }
    if (pid > 0) {
	/* parent */
        comm_close(cfd);        /* close shared socket with child */
	sfd = comm_open(SOCK_STREAM, 0, local_addr, 0, 0, NULL);
        if (sfd == COMM_ERROR)
            return -1;
        if (comm_connect_addr(sfd, &S) == COMM_ERROR) {
            comm_close(sfd);
            return -1;
        }
        if (write(sfd, "$hello\n", 7) < 0) {
            debug(SECTION, 0, "softForkRecoder: $hello write test failed\n");
            comm_close(sfd);
            return -1;
        }
        memset(buf, '\0', 128);
        if (read(sfd, buf, 127) < 0) {
            debug(50, 0, "softForkRecoder: $hello read test failed\n");
            debug(50, 0, "--> read: %s\n", xstrerror());
            comm_close(sfd);
            return -1;
        } else if (strcmp(buf, "$alive\n")) {
            debug(50, 0, "softForkRecoder: $hello read test failed\n");
            debug(50, 0, "--> got: '%s'\n", rfc1738_escape(buf));
            comm_close(sfd);
            return -1;
        }
        comm_set_fd_lifetime(sfd, -1);
        return sfd;
    }
    /* child */
    no_suid();
    if ((fd = accept(cfd, NULL, NULL)) < 0) {
        debug(50, 0, "softForkRecoder: FD %d accept: %s\n", cfd, xstrerror());
        _exit(1);
    }
    if (dup2(fd, 0) < 0) {
        debug(50, 0, "softForkRecoder: FD 0 dup2: %s\n", xstrerror());
        _exit(1);
    }
    if (dup2(fd, 1) < 0) {
        debug(50, 0, "softForkRecoder: FD 1 dup2: %s\n", xstrerror());
        _exit(1);
    }
    if (dup2(fileno(debug_log), 2) < 0) {
        debug(50, 0, "softForkRecoder: FD 2 dup2: %s\n", xstrerror());
        _exit(1);
    }
    fclose(debug_log);
    close(fd);
    close(cfd);
    execlp(prg, "(recoder)", (char *)NULL);
    
    /* should not happen */
    debug(50, 0, "softForkRecoder: %s: %s\n", prg, xstrerror());
    _exit(1);
    return -1;
}

void
softFreeMemory(void)
{
    int i;
    recoderequest_t *r1, *r2;

    if (recoder_table) {
	for (i = 0; i < NRecodersAlloc; i++) {
	    if (recoder_table[i]->list->next) {
		r1 = recoder_table[i]->list;
		r2 = r1->next;
		while (r2) {
		    safe_free(r1);
		    r1 = r2;
		    r2 = r1->next;
		}
		safe_free(r1);
	    } else {
		safe_free(recoder_table[i]->list);
	    }
	    safe_free(recoder_table[i]);
	}
	safe_free(recoder_table);
    }
}

void
softStartRecoders(void)
{
    char *prg = Config.Program.recoder;
    int N = Config.recoderChildren;
    int i, rd;
    LOCAL_ARRAY(char, fd_note_buf, FD_ASCII_NOTE_SZ);

    softFreeMemory();
    recoder_table = xcalloc(N, sizeof(softrecoder_t *));
    NRecodersAlloc = 0;

    debug(SECTION, 5, "Configured with %d recoders\n", N);
    debug(SECTION, 5, "Recoder is %s\n", prg);
    
    for (i = 0; i < N; i++) {
	recoder_table[i] = xcalloc(1, sizeof(softrecoder_t));
	if ((rd = softForkRecoder(prg)) < 0) {
	    debug(SECTION, 1,
		  "softStartRecoders: Failed to start 'recoder' #%d.\n", i+1);
	    recoder_table[i]->flags &= ~SOFT_FLAG_ALIVE;
	    recoder_table[i]->id = i + 1;
	    recoder_table[i]->inpipe = -1;
	    recoder_table[i]->outpipe = -1;
	} else {
	   debug(SECTION, 4,
		  "softStartRecoders: FD %d connected to %s #%d.\n", 
		 rd, prg, i+1);
	   recoder_table[i]->flags |= SOFT_FLAG_ALIVE;
	   recoder_table[i]->id = i + 1;
	   recoder_table[i]->inpipe = rd;
	   recoder_table[i]->outpipe = rd;
	   recoder_table[i]->fin = fdopen(rd, "r");
	   recoder_table[i]->answer = squid_curtime;
	   recoder_table[i]->dispatch_time = current_time;
	   recoder_table[i]->list = xcalloc(1, sizeof(recoderequest_t));
	   recoder_table[i]->list->next = NULL;
	   recoder_table[i]->list->e = NULL;
	   recoder_table[i]->list->slices = -1;
	   sprintf(fd_note_buf, "%s #%d", prg, recoder_table[i]->id);
	   fd_note(recoder_table[i]->inpipe, fd_note_buf);
	   commSetNonBlocking(recoder_table[i]->inpipe);
	   debug(SECTION, 3,
		 "softStartRecoders: 'recoder' %d started.\n", i + 1);
	   NRecodersAlloc++;
	}
    }
    if (NRecodersAlloc == 0 && Config.recoderChildren > 0)
	fatal("Failed to start any recoders");

    debug(SECTION, 1, "Started %d 'recoder' processes\n", NRecodersAlloc);
}

void
softShutdownRecoders(void)
{
    int i, l;
    LOCAL_ARRAY(char, com, 16);

    sprintf(com, "$shutdown\n");
    l = strlen(com);

    for (i = 0; i < NRecodersAlloc; i++) {
	if (recoder_table[i]->flags & SOFT_FLAG_CLOSING)
	    continue;

	if(write(recoder_table[i]->outpipe, com, l) != l) {
	    debug(45, 1, "softShutdownRecoders: write: %s\n", xstrerror());
	}
	recoder_table[i]->flags |= SOFT_FLAG_CLOSING;
	/*recoder_table[i]->flags &= ~SOFT_FLAG_ALIVE;*/
    }
}

softrecoder_t *
softGetNextRecoder(void)
{
    int i, min, n, tmp;
    
    min = recoder_table[0]->waiting + recoder_table[0]->pending;
    n = 0;

    for (i = 0; i < NRecodersAlloc; i++) {
	if (recoder_table[i]->waiting == 0 &&
	    recoder_table[i]->pending == 0 &&
	    !(recoder_table[i]->flags & SOFT_FLAG_BUSY) &&
	    !(recoder_table[i]->flags & SOFT_FLAG_CLOSING)) {
	    return recoder_table[i];
	}
	tmp = recoder_table[i]->waiting + recoder_table[i]->pending;
	if (tmp < min) {
	    min = tmp;
	    n = i;
	}
    }
    return recoder_table[n];
}

int
softInsertRequest(softrecoder_t *recoder, StoreEntry *e, int slices)
{
    recoderequest_t *tmp;
    recoderequest_t *req = xcalloc(1, sizeof(recoderequest_t));
    int n = 0;

    req->e = e;
    req->slices = slices;
    req->flags = REQUEST_FLAG_PENDING;
    req->next = (recoderequest_t *)NULL;
    
    tmp = recoder->list;
    
    while (tmp->next) {
	tmp = tmp->next;
	n++;
    }

    tmp->next = req;
    recoder->pending++;
    return n;
}

recoderequest_t *
softGetNextPending(softrecoder_t *recoder)
{
    recoderequest_t *req;
    
    req = recoder->list;
    while ((req = req->next)) {
	if (req->flags & REQUEST_FLAG_PENDING) {
	    break;
	}
    }
    return req;
}

void
softDoWrite(int fd, softrecoder_t *recoder)
{
    int l;
    recoderequest_t *req;
    LOCAL_ARRAY(char, command, SOFT_COMBUF_SZ);
    LOCAL_ARRAY(char, file, SQUID_MAXPATHLEN);

    if (!recoder) {
	debug(SECTION, 0, "softDoWrite: No Recoder!\n");
    }
    if (fd != recoder->outpipe) {
	debug(SECTION, 0, "softDoWrite: FD %d and outpipe %d!\n", fd, 
	      recoder->outpipe);
    }

    debug(SECTION, 5, "softDoWrite: writing for recoder %d\n", recoder->id);

    if (recoder->pending <= 0) {
	debug(SECTION, 0, "softDoWrite: no requests pending at %d!\n", 
	      recoder->id);
	return;
    }
    if ((req = softGetNextPending(recoder)) == NULL) {
	debug(SECTION, 0, "softDoWrite: NULL request at %d?\n", recoder->id);
	return;
    }

    storeSwapFullPath((req->e)->swap_file_number, file);
    sprintf(command, "%s %d\n", file, req->slices);
    l = strlen(command);

    store_swap_size -= ((req->e->object_len / 10) + 1023) >> 10;

    debug(SECTION, 7, "softDoWrite: command is %s\n", command);
    if (write(recoder->outpipe, command, l) != l) {
	debug(SECTION, 1, "softDoWrite: %s\n", xstrerror());
	debug(SECTION, 1, "  --> writing to %d\n", recoder->id);
	return;
    }
    req->flags |= REQUEST_FLAG_WRITTEN;
    req->flags &= ~REQUEST_FLAG_PENDING;
    req->dispatched = current_time;

    recoder->pending--;
    recoder->waiting++;
    recoder->flags |= SOFT_FLAG_BUSY;
    recoder->dispatch_time = current_time;
    SoftStats.requests++;

    if (recoder->pending <= 0) {
	recoder->flags &= ~SOFT_FLAG_WRITE_PENDING;
    } else {
	commSetSelect(recoder->outpipe, COMM_SELECT_WRITE, (PF) softDoWrite,
		      recoder, 0);
    }
    commSetSelect(recoder->inpipe, COMM_SELECT_READ, (PF) softGetRecodedResult,
		  recoder, 0);
    return;
}

void
softGetRecodedResult(int fd, softrecoder_t *recoder)
{
    int newlen, newslices, converted, svc_time, total_time, recode_time;
    LOCAL_ARRAY(char, combuf, SOFT_COMBUF_SZ);
    recoderequest_t *req;

    if (recoder->waiting <= 0) {
	debug(SECTION, 0, "softGetRecodedResult: no requests written to %d\n",
		recoder->id);
	return;
    }

    memset(combuf, '\0', SOFT_COMBUF_SZ);
    if (fgets(combuf, SOFT_COMBUF_SZ, recoder->fin) == NULL) {
	debug(SECTION, 1, "softGetRecodedResults: fgets: %s\n", xstrerror());
	return;
    }

    recoder->waiting--;
    SoftStats.replies++;
    req = recoder->list->next;
    svc_time = tvSubMsec(req->dispatched, current_time);
    total_time = (int)(recoder->completed * recoder->avg_servicetime);
    recoder->completed++;
    recoder->avg_servicetime = (float)((total_time + svc_time) / 
				      recoder->completed);

    if (recoder->waiting == 0) {
	recoder->flags &= ~SOFT_FLAG_BUSY;
	if (recoder->pending == 0) {
	    recoder->list->next = (recoderequest_t *) NULL;
	} else {
	    recoder->list->next = req->next;
	}
    } else {
	recoder->list->next = req->next;
	commSetSelect(recoder->inpipe, COMM_SELECT_READ, 
		      (PF) softGetRecodedResult, recoder, 0);
    }
    sscanf(combuf, "%d %d %d %d\n", &newlen, &newslices, &converted, 
	   &recode_time);
    total_time = (int)((recoder->completed - 1) * recoder->avg_recodetime);
    recoder->avg_recodetime = (float)((total_time + recode_time) /
				      recoder->completed);
    debug(SECTION, 4, "softGetRecodedResults: %s\n", req->e->key);
    debug(SECTION, 4, "     New length: %d   New slices: %d   Converted: %d\n",
	  newlen, newslices, converted);
    softUpdate(req->e, newlen, newslices, converted);
    safe_free(req);
}

/* FC #modif# the following function has been added by me */
/* recode an object in the disk cache */
/* return number of objects recoded. */
/* YG: add entry for the case of wrong image */
int
storeSoftRecode(StoreEntry * e, int justconvert)
{
    StoreEntry *result = NULL;
    StoreEntry *hentry = NULL;
    hash_link *hptr = NULL;
    const char *hkey;
    
    debug(20, 3, "storeSoftRecode: Recoding: '%s'\n", e->key);
    /* if the object is 'recoding' do nothing */
    if(e->soft_status == RECODING) {
	debug(20, 0, "storeSoftRecode: Recoding a 'RECODING' object\n");
	return 0;
    }
    /*debug(SECTION, 6, "storeSoftRecode: 1 for %s\n", e->key);*/
    /* I want to be sure not to recode minimum information objects */
    if(e->soft_slices == 1) {
	debug(20, 0, 
	      "storeSoftRecode: Recoding a minimum information object\n");
	return -storeRelease(e);
    }

    /* If, for any reason we can't discard this object because of an
     * outstanding request, mark it for pending release */
     if (storeEntryLocked(e)) {
	storeExpireNow(e);
	debug(20, 3, "storeSoftRecode: Only setting RELEASE_REQUEST bit\n");
	storeReleaseRequest(e);
	return 0;
    }
    if (e->key != NULL) {
        if(e->mem_status != NOT_IN_MEMORY) {
	    debug(20, 0,
		  "storeSoftRecode: '%s' in memory, cannot recode it!\n");
	    return 0;
	}
	if ((hptr = hash_lookup(store_table, e->key)) == NULL) {
	    debug(20, 0, "storeSoftRecode: Not Found: '%s'\n", e->key);
	    debug(20, 0, "Dump of Entry 'e':\n %s\n", storeToString(e));
	    debug_trap("storeSoftRecode: Invalid Entry");
	    return 0;
	}
	result = (StoreEntry *) hptr;
	if (result != e) {
	    debug(20, 0, "storeSoftRecode: Duplicated entry? '%s'\n",
		result->url ? result->url : "NULL");
	    debug(20, 0, "Dump of Entry 'e':\n%s", storeToString(e));
	    debug(20, 0, "Dump of Entry 'result':\n%s",
		  storeToString(result));
	    debug_trap("storeSoftRecode: Duplicate Entry");
	    return 0;
	}
    }
    /* check if coresponding HEAD object exists. */
    if (e->method == METHOD_GET) {
	hkey = storeGeneratePublicKey(e->url, METHOD_HEAD);
	if ((hentry = (StoreEntry *) hash_lookup(store_table, hkey)))
	    storeExpireNow(hentry);
    }
    if (store_rebuilding == STORE_REBUILDING_FAST) {
	debug(20, 2,
	      "storeSoftRecode: Delaying recode until store is rebuilt: '%s'\n",
	      e->key ? e->key : e->url ? e->url : "NO URL");
	storeExpireNow(e);
	storeSetPrivateKey(e);
	return 0;
    }
    if (e->key)
	debug(20, 5, "storeSoftRecode: Recode object key: %s\n", e->key);
    else
	debug(20, 5, "storeSoftRecode: Recode anonymous object\n");

    /* call the recoder */
    return softRecode(e, justconvert);
}

void
softSetNewStatus(StoreEntry *e, int converted)
{
    if (!BIT_TEST(e->soft_flag, SOFT_PROCESSED)) {
	/* First time this image is processed */
	e->soft_status = CONVERTED;
	switch (e->soft_slices) {
	case 5: 
	    e->soft_status = RECODED_OK;
	    /* FALLTHROUGH */
	case 6: {
	    BIT_SET(e->soft_flag, SOFT_SIXLEVEL);
	    break;
	}
	case 9:
	    e->soft_status = RECODED_OK;
	    /* FALLTHROUGH */
	case 10:{
	    BIT_SET(e->soft_flag, SOFT_TENLEVEL);
	    break;
	}
	default:{
	    BIT_SET(e->soft_flag, SOFT_OTHERLEVEL);
	    break;
	}
	}
	if (converted) {
	    BIT_RESET(e->soft_flag, SOFT_ORIGINAL_PROGRESSIVE);
	} else {
	    BIT_SET(e->soft_flag, SOFT_ORIGINAL_PROGRESSIVE);
	}
	BIT_SET(e->soft_flag, SOFT_PROCESSED);
    } else {
	e->soft_status = RECODED_OK;
    }
}

/* Handler to be called when an image has been recoded */
void 
softUpdate(StoreEntry *e, int newlen, int newslices, int converted)
{
    store_swap_size += ((e->object_len / 10) + 1023) >> 10;
    
    if(e->mem_status != NOT_IN_MEMORY) {
	debug(SECTION, 1,
	      "softUpdate: object accessed during recoding: %s\n",
	      e->url);
    }
    if(newlen == 0 || newslices <= 0) {
	/* Error during recoding, release the image. */
	storeRelease(e);
    } else {
	/* everything ok, object succesfully recoded */
	store_swap_size -= (e->object_len + 1023) >> 10;
	e->object_len = newlen;
	store_swap_size += (e->object_len + 1023) >> 10;
	/* make last reference newer, so the same object is not re-recoded
	 * suddenly */
	e->lastref = current_time.tv_sec;
	e->soft_slices = newslices;
	softSetNewStatus(e, converted);
	/*e->soft_status = RECODED_OK;*/
	storeLog(STORE_LOG_RECODEEND, e);
    }
}

int
softHowManySlices(StoreEntry *e) 
{
    return 1;
}

/* Hand image over to the recoder to be recoded. */
int 
softRecode(StoreEntry * e, int justconvert)
{
    int howmuch;
    softrecoder_t *recoder;

    debug(SECTION, 3, "softRecode: recode '%s'\n", e->key);
    /* REMOVE THIS */
    debug(72, 0, "Recoding %s\n", e->url ? e->url : e->key);

    recoder = softGetNextRecoder();

    if(recoder->outpipe < 0) {
	debug_trap("softRecode: recoder outpipe < 0");
	return 0;
    }

    howmuch = justconvert ? 0 : softHowManySlices(e);
    softInsertRequest(recoder, e, howmuch);
    SoftStats.hist[recoder->id - 1]++;

    commSetSelect(recoder->outpipe, COMM_SELECT_WRITE, (PF) softDoWrite,
		  recoder, 0);

    debug(SECTION, 5, "softRecode: Dispatched recode for %s\n",
	  e->key != NULL ? e->key : "(no key)");
    debug(SECTION, 5, "     Recoder: %d     Slices: %d\n", recoder->id,
	  howmuch);
    e->soft_status = RECODING;
    /* log the action */
    storeLog(STORE_LOG_RECODESTART, e);
    return 1;
}

/* Shutdown soft. */
void 
softClose(void)
{
    /* softShutdownRecoders(); */
    softFreeMemory();
    return;
}

void
softStats(StoreEntry *sentry)
{
    int i;
    int total_svc = 0;
    int total_rcd = 0;
    int requests = 0;
    softrecoder_t *recoder = NULL;

    storeAppendPrintf(sentry, "{Recoder Statistics:\n");
    storeAppendPrintf(sentry, "{recoder requests: %d}\n", SoftStats.requests);
    storeAppendPrintf(sentry, "{recoder replies: %d}\n", SoftStats.replies);
    storeAppendPrintf(sentry, "{number of recoders: %d}\n", NRecodersAlloc);
    storeAppendPrintf(sentry, "{recoder use histogram:}\n");
    for (i = 0; i < NRecodersAlloc; i++) {
	storeAppendPrintf(sentry, "{    recoder #%d: %d}\n", i + 1, 
			  SoftStats.hist[i]);
	requests += recoder_table[i]->completed;
	total_svc += (int)(recoder_table[i]->completed * 
			   recoder_table[i]->avg_servicetime);
	total_rcd += (int)(recoder_table[i]->completed * 
			   recoder_table[i]->avg_recodetime);
    }
    if (requests) {
	storeAppendPrintf(sentry, "{Average service time: %.2f ms}\n",
			  (float)(total_svc / requests));
	storeAppendPrintf(sentry, "{Average recode time: %.2f ms}\n",
			  (float)(total_rcd / requests));
    } else {
	storeAppendPrintf(sentry, "{No recode requests served.}\n");
    }
    storeAppendPrintf(sentry, "{}\n");
    storeAppendPrintf(sentry, "{recoders status:}\n");
    for (i = 0; i < NRecodersAlloc; i++) {
	recoder = *(recoder_table + i);
	storeAppendPrintf(sentry, "{recoder #%d:}\n", i + 1);
	storeAppendPrintf(sentry, "{    Flags: %c%c%c\n}\n",
			  recoder->flags & SOFT_FLAG_ALIVE ? 'A' : ' ',
			  recoder->flags & SOFT_FLAG_BUSY ? 'B' : ' ',
			  recoder->flags & SOFT_FLAG_CLOSING ? 'C' : ' ');
			  
	storeAppendPrintf(sentry, "{    FDs (in/out): %d/%d}\n",
			  recoder->inpipe, recoder->outpipe);
	storeAppendPrintf(sentry, "{    Alive since: %d}\n",
			  mkrfc1123(recoder->answer));
	storeAppendPrintf(sentry, "{    Last Dispatched: %0.3f second ago}\n",
			  0.001 * tvSubMsec(recoder->dispatch_time, 
					    current_time));
	storeAppendPrintf(sentry, "{    Pending requests (write): %d}\n",
			  recoder->pending);
	storeAppendPrintf(sentry, "{    Waiting requests (recode): %d}\n",
			  recoder->waiting);
	storeAppendPrintf(sentry, "{    Average service time: %.2f ms}\n",
			  recoder->avg_servicetime);
	storeAppendPrintf(sentry, "{    Average recode time: %.2f ms}\n",
			  recoder->avg_recodetime);
    } 
    
    storeAppendPrintf(sentry, "\n{Flags key:}\n\n");
    storeAppendPrintf(sentry, "{   A = ALIVE}\n");
    storeAppendPrintf(sentry, "{   B = BUSY}\n");
    storeAppendPrintf(sentry, "{   C = CLOSING}\n");
    
    storeAppendPrintf(sentry, close_bracket);
}

#endif /* _SOFT_CACH_ */
